import io
import cv2
import base64
import logging
import numpy as np
from PIL import Image
from pynput import keyboard
from eymos import Service, log


class DependencyLoader:
	TKINTER = False

	@staticmethod
	def is_tkinter():
		"""Check if tkinter is installed."""
		return DependencyLoader.TKINTER


try:
	import tkinter as tk
	from tkinter import Canvas
	from PIL import ImageTk
	DependencyLoader.TKINTER = True
except ImportError:
	pass


class ImageSize:
	COVER = 0
	CONTAIN = 1
	STRETCH = 2


class WindowService(Service):
	DEFAULT_FPS = 60
	DEFAULT_RESOLUTION = [480, 270]
	APP_LOGO = " "

	def __init__(self, name: str, config: dict, services: dict):
		"""Initialize the service.
		Args:
			name (str): The name of the service.
			config (dict): The system configuration.
			services (dict, optional): The services to use. Defaults to {}.
		"""
		# Initialize the service attributes
		self.__fps = None
		self.__title = None
		self.__resizable = None
		self.__always_on_top = None
		self.__background = None
		self.__border = None
		self.__width = None
		self.__height = None
		self.__draggable = None
		self.__toggle_visibility = None
		self.__frame = None
		self.__tk_root = None
		self.__tk_icon = None
		self.__tk_image = None
		self.__tk_photo = None
		self.__tk_canvas = None
		self.__tk_canvas_img = None
		self.__tk_is_visible = None

		# Call the parent class constructor
		super().__init__(name, config, services)

	def init(self):
		"""Initialize the service."""
		self.__fps = self._config.get('fps', self.DEFAULT_FPS)
		self._loop_delay = 1 / self.__fps
		self.__title = self._config.get('title', 'EymOS')
		self.__resizable = self._config.get('resizable', False)
		self.__always_on_top = self._config.get('always_on_top', True)
		self.__background = self._config.get('background', 'black')
		self.__border = self._config.get('border', False)
		self.__width = self._config.get('width', self.DEFAULT_RESOLUTION[0])
		self.__height = self._config.get('height', self.DEFAULT_RESOLUTION[1])
		self.__draggable = self._config.get('draggable', True)
		self.__toggle_visibility = self._config.get('toggle_visibility', {'key': 'h', 'modifiers': ['ctrl']})
		self.__frame = 0
		self.__tk_root = None
		self.__tk_icon = self.APP_LOGO
		self.__tk_image = None
		self.__tk_photo = None
		self.__tk_canvas = None
		self.__tk_canvas_img = None
		self.__tk_is_visible = True

		# Initialize the tkinter window
		self.__tk_init()

	def destroy(self):
		"""Destroy the service."""
		self.__fps = None
		self.__title = None
		self.__resizable = None
		self.__always_on_top = None
		self.__background = None
		self.__border = None
		self.__width = None
		self.__height = None
		self.__draggable = None
		self.__toggle_visibility = None
		self.__frame = None
		self.__tk_icon = None
		self.__tk_image = None
		self.__tk_photo = None
		self.__tk_canvas = None
		self.__tk_canvas_img = None
		self.__tk_is_visible = None

		# Destroy the tkinter window from manager
		if self._manager:
			self._manager.on_stop(self.__tk_root.quit)

	def loop(self):
		"""Service loop."""
		if not DependencyLoader.is_tkinter():
			return

		# Update the tkinter window
		if self.__tk_root is not None:
			self.__tk_update()

	def mainloop(self):
		"""Start the tkinter main loop."""
		# Check if tkinter is installed
		if self.__tk_root is None:
			return

		# Start the tkinter main loop
		self.__tk_root.mainloop()

	def draw(self, image: np.ndarray, size: int = ImageSize.COVER):
		"""Draw the image on the window.
		Args:
			image (np.ndarray): The image to draw.
		"""
		# Draw the image on the tkinter canvas
		self.__tk_draw(image, size)

	def __tk_init(self):
		"""Initializes tkinter for PCs to emulate the screen."""
		# Check if tkinter is installed
		if not DependencyLoader.is_tkinter():
			log('Tkinter is not installed. The window service is disabled.', logging.WARNING)
			return

		# Create the tkinter window
		geometry = f'{self.__width}x{self.__height}'
		self.__tk_root = tk.Tk()
		self.__tk_root.title(self.__title)
		self.__tk_root.geometry(geometry)
		self.__tk_root.resizable(self.__resizable, self.__resizable)

		# Set the window always on top
		if self.__always_on_top:
			self.__tk_root.lift()

		# Remove the window border
		if not self.__border:
			self.__tk_root.overrideredirect(True)
			self.__tk_root.config(bg=self.__background)

		# Create the tkinter canvas
		self.__tk_image = Image.new('RGB', (self.__width, self.__height), 'black')
		self.__tk_photo = ImageTk.PhotoImage(self.__tk_image)
		self.__tk_canvas = Canvas(self.__tk_root, width=self.__width, height=self.__height, bg=self.__background, highlightthickness=0)
		self.__tk_canvas_img = self.__tk_canvas.create_image(0, 0, anchor='nw', image=self.__tk_photo)
		self.__tk_canvas.pack()

		# Set the window icon
		if self.__tk_icon:
			icon_data = base64.b64decode(self.__tk_icon)
			icon_image = Image.open(io.BytesIO(icon_data))
			self.__tk_icon = ImageTk.PhotoImage(icon_image)
			self.__tk_root.iconphoto(False, self.__tk_icon)

		# Enable dragging
		if self.__draggable:
			self.__tk_drag(self.__tk_root)

		# Enable visibility toggle
		if self.__toggle_visibility:
			self.__tk_visibility(self.__tk_root)

	def __tk_drag(self, root):
		"""Enable dragging for the tkinter window.
		Args:
			root (tk.Tk): The tkinter window.
		"""
		# Check if tkinter is installed
		if not DependencyLoader.is_tkinter():
			return

		# Initialize the start coordinates
		coords = {"x": 0, "y": 0}

		# Function to start dragging
		def start_drag(event, coords):
			coords["x"] = event.x
			coords["y"] = event.y

		# Function to stop dragging
		def stop_drag(event, coords):
			coords["x"] = None
			coords["y"] = None

		# Function to drag the window
		def drag(event, coords):
			x = root.winfo_x() + event.x - coords["x"]
			y = root.winfo_y() + event.y - coords["y"]
			root.geometry(f'+{x}+{y}')

		# Remove the window border
		if not self.__border:
			self.__tk_root.overrideredirect(True)

		# Bind the mouse events
		root.bind('<ButtonPress-1>', lambda event: start_drag(event, coords))
		root.bind('<ButtonRelease-1>', lambda event: stop_drag(event, coords))
		root.bind('<B1-Motion>', lambda event: drag(event, coords))

	def __tk_visibility(self, root):
		"""Enable visibility toggle for the tkinter window.
		Args:
			root (tk.Tk): The tkinter window.
		"""
		# Check if tkinter is installed
		if not DependencyLoader.is_tkinter():
			return

		# Check if the toggle key is set
		if not self.__toggle_visibility:
			return

		# Initialize the modifier keys
		modifiers = self.__toggle_visibility.get('modifiers', [])
		key_char = self.__toggle_visibility.get('key', 'h')

		# Initialize the pressed keys
		keys = {"ctrl": False, "shift": False, "alt": False, "meta": False}

		# Function to check pressed keys
		def on_press(key, keys):
			if keyboard.Key.ctrl_l == key or keyboard.Key.ctrl_r == key:
				keys["ctrl"] = True
			elif keyboard.Key.shift_l == key or keyboard.Key.shift_r == key:
				keys["shift"] = True
			elif keyboard.Key.alt_l == key or keyboard.Key.alt_r == key:
				keys["alt"] = True
			elif keyboard.Key.cmd_l == key or keyboard.Key.cmd_r == key:
				keys["meta"] = True
			is_pressed = True
			for modifier in modifiers:
				if not keys[modifier]:
					is_pressed = False
					break
			if type(key) != keyboard.KeyCode:
				return
			if is_pressed and key_char == key.char:
				self.__tk_is_visible = not self.__tk_is_visible
				if self.__tk_is_visible:
					log('The window is visible')
					root.deiconify()
				else:
					log('The window is hidden')
					root.withdraw()

		# Function to check released keys
		def on_release(key, keys):
			if keyboard.Key.ctrl_l == key or keyboard.Key.ctrl_r == key:
				keys["ctrl"] = False
			elif keyboard.Key.shift_l == key or keyboard.Key.shift_r == key:
				keys["shift"] = False
			elif keyboard.Key.alt_l == key or keyboard.Key.alt_r == key:
				keys["alt"] = False
			elif keyboard.Key.cmd_l == key or keyboard.Key.cmd_r == key:
				keys["meta"] = False

		# Start the listener
		listener = keyboard.Listener(on_press=lambda key: on_press(key, keys), on_release=lambda key: on_release(key, keys))
		listener.start()

	def __tk_update(self):
		"""Update the tkinter window."""
		# Check if the window is visible or the canvas is not set
		if not self.__tk_is_visible or self.__tk_canvas is None:
			return

		# Update the frame
		self.__frame += 1

		# Update width and height
		width = self.__tk_root.winfo_width()
		height = self.__tk_root.winfo_height()
		if width != self.__width:
			self.__width = width
		if height != self.__height:
			self.__height = height

		# Set the always on top
		if self.__always_on_top:
			self.__tk_root.attributes('-topmost', True)

		# Draw the image
		self.__tk_draw()

		# Update the window
		self.__tk_root.update_idletasks()
		self.__tk_root.update()

	def __tk_clear(self):
		"""Clear the tkinter canvas."""
		# Draw a black image
		self.__tk_image = Image.new('RGB', (self.__width, self.__height), 'black')

		# Convert the image to tkinter
		self.__tk_photo = ImageTk.PhotoImage(self.__tk_image)

		# Update the canvas
		self.__tk_canvas.itemconfig(self.__tk_canvas_img, image=self.__tk_photo)

	def __tk_draw(self, image: np.ndarray = None, size: int = ImageSize.COVER):
		"""Draw the image on the tkinter canvas.
		Args:
			image (Image, optional): The image to draw. Defaults to None.
		"""
		# Check if the image is set
		if image is None:
			return

		# Resize the image
		if size == ImageSize.COVER:
			image = self.__image_resize_cover(image, self.__width, self.__height)
		elif size == ImageSize.CONTAIN:
			image = self.__image_resize_contain(image, self.__width, self.__height)
		elif size == ImageSize.STRETCH:
			image = self.__image_resize_stretch(image, self.__width, self.__height)

		# Convert the image to PIL
		image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
		image = Image.fromarray(image)

		# Convert the image to tkinter
		self.__tk_photo = ImageTk.PhotoImage(image)

		# Update the canvas
		self.__tk_canvas.itemconfig(self.__tk_canvas_img, image=self.__tk_photo)

	@staticmethod
	def __image_resize_cover(image: np.ndarray, width: int, height: int) -> np.ndarray:
		"""Resize the image to cover the window.
		Args:
			image (np.ndarray): The image to resize.
			width (int): The width of the window.
			height (int): The height of the window.
		Returns:
			np.ndarray: The resized image.
		"""
		# Get the image size
		image_width, image_height = image.shape[1], image.shape[0]

		# Calculate the resize ratio
		ratio = max(width / image_width, height / image_height)

		# Resize the image
		image = cv2.resize(image, (int(image_width * ratio), int(image_height * ratio)))

		# Calculate the cropping
		crop_x = (image.shape[1] - width) // 2
		crop_y = (image.shape[0] - height) // 2

		# Crop the image
		image = image[crop_y:crop_y + height, crop_x:crop_x + width]

		# Return the image
		return image

	@staticmethod
	def __image_resize_contain(image: np.ndarray, width: int, height: int) -> np.ndarray:
		"""Resize the image to contain the window.
		Args:
			image (np.ndarray): The image to resize.
			width (int): The width of the window.
			height (int): The height of the window.
		Returns:
			np.ndarray: The resized image.
		"""
		# Get the image size
		image_width, image_height = image.shape[1], image.shape[0]

		# Calculate the resize ratio
		ratio = min(width / image_width, height / image_height)

		# Resize the image
		image = cv2.resize(image, (int(image_width * ratio), int(image_height * ratio)))

		# Calculate the padding
		padding = (height - image.shape[0]) // 2

		# Add the padding
		image = cv2.copyMakeBorder(image, padding, padding, 0, 0, cv2.BORDER_CONSTANT, value=[0, 0, 0])

		# Return the image
		return image

	@staticmethod
	def __image_resize_stretch(image: np.ndarray, width: int, height: int) -> np.ndarray:
		"""Resize the image to stretch the window.
		Args:
			image (np.ndarray): The image to resize.
			width (int): The width of the window.
			height (int): The height of the window.
		Returns:
			np.ndarray: The resized image.
		"""
		# Resize the image
		image = cv2.resize(image, (width, height))

		# Return the image
		return image



import cv2
import base64
from PIL import Image
from eymos.service import Service


class FrameType:
	IMAGE = 0    # PIL Image
	NUMPY = 1    # Numpy array
	LIST = 2     # List
	BYTES = 3    # Bytes
	BASE64 = 4   # Base64 string


class CameraService(Service):
	DEFAULT_FPS = 25
	DEFAULT_RESOLUTION = [640, 480]

	def __init__(self, name: str, config: dict, services: dict):
		"""Initialize the service.
		Args:
			name (str): The name of the service.
			config (dict): The system configuration.
			services (dict, optional): The services to use. Defaults to {}.
		"""
		# Initialize the service attributes
		self.__camera = None
		self.__last_frame = None
		self.__last_frame_list = None
		self.__last_frame_bytes = None
		self.__last_frame_image = None
		self.__last_frame_base64 = None
		self.__rotate_frame = None
		self.__resolution = None
		self.__fps = None

		# Call the parent class constructor
		super().__init__(name, config, services)

	def init(self):
		"""Initialize the service."""
		self.__camera = None
		self.__last_frame = None
		self.__last_frame_list = None
		self.__last_frame_bytes = None
		self.__last_frame_image = None
		self.__last_frame_base64 = None
		self.__rotate_frame = self._config.get('rotate_frame', False)
		self.__resolution = self._config.get('resolution', self.DEFAULT_RESOLUTION)
		self.__fps = self._config.get('fps', self.DEFAULT_FPS)
		self._loop_delay = 1 / self.__fps

	def destroy(self):
		"""Destroy the service."""
		if self.__camera is not None:
			self.__camera.release()
			self.__camera = None
		self.__last_frame = None
		self.__last_frame_list = None
		self.__last_frame_bytes = None
		self.__last_frame_image = None
		self.__last_frame_base64 = None
		self.__rotate_frame = None
		self.__resolution = None
		self.__fps = None
		self._loop_delay = self.LOOP_DELAY

	def before(self):
		"""Before the loop. (Before the loop method is called, in the service thread)"""
		# Connect to the camera
		self.__camera = cv2.VideoCapture(self._config.get('camera', 0))

		# Check if the camera is opened
		if not self.__camera.isOpened():
			raise RuntimeError("Could not open camera")

	def loop(self):
		"""Service loop."""
		# Reset the last frame. This optimizes the memory usage by reusing the last frame.
		self.__last_frame = None
		self.__last_frame_list = None
		self.__last_frame_bytes = None
		self.__last_frame_image = None
		self.__last_frame_base64 = None

	def __convert_frame(self, frame_type: int):
		"""Convert the camera frame.
		Args:
			frame_type (int): The frame type.
		Returns:
			Any: The converted frame, or None if the frame type is invalid.
		"""
		# Get the last frame and check if it exists
		frame = self.__last_frame
		if frame is None:
			return None

		# Do not convert the frame if it is in numpy format
		if frame_type == FrameType.NUMPY:
			return frame

		# Convert the frame to a list, if needed
		if frame_type == FrameType.LIST:
			if self.__last_frame_list is None:
				self.__last_frame_list = frame.tolist()
			return self.__last_frame_list

		# Convert the frame to bytes, if needed
		if frame_type == FrameType.BYTES:
			if self.__last_frame_bytes is None:
				self.__last_frame_bytes = cv2.imencode('.jpg', frame)[1].tobytes()
			return self.__last_frame_bytes

		# Convert the frame to a PIL image, if needed
		if frame_type == FrameType.IMAGE:
			frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
			if self.__last_frame_image is None:
				self.__last_frame_image = Image.fromarray(frame)
			return self.__last_frame_image

		# Convert the frame to a base64 string, if needed
		if frame_type == FrameType.BASE64:
			if self.__last_frame_base64 is None:
				retval, buffer = cv2.imencode('.png', frame)
				self.__last_frame_base64 = base64.b64encode(buffer).decode('utf-8')
			return self.__last_frame_base64

		# Return None if the frame type is invalid
		return None

	def get_frame(self, frame_type: int = FrameType.NUMPY):
		"""Get the camera frame.
		Args:
			frame_type (int, optional): The frame type. Defaults to FrameType.NUMPY.
		Returns:
			Any: The camera frame, or None if the frame type is invalid.
		"""
		# Get the last frame
		frame = self.__last_frame

		# Return the converted frame if it exists
		if frame is not None:
			return self.__convert_frame(frame_type)

		# Return None if the camera is not connected
		if self.__camera is None:
			return None

		# Read the frame from the camera
		ret, frame = self.__camera.read()

		# Return None if the frame could not be read
		if not ret:
			return None

		# Resize and rotate the frame
		frame = cv2.resize(frame, tuple(self.__resolution))

		# Rotate the frame if needed
		if self.__rotate_frame:
			frame = cv2.rotate(frame, cv2.ROTATE_180)

		# Save the last frame
		self.__last_frame = frame

		# Return the converted frame
		return self.__convert_frame(frame_type)



import abc
import time
import logging
import threading

from .logger import log


class Service(metaclass=abc.ABCMeta):
	DEPENDENCIES = []
	MAX_ERRORS = 5
	MAX_ERRORS_REBOOT = -1
	ERROR_INTERVAL = 5
	LOOP_DELAY = 0.25

	def __init__(self, name: str, config: dict, services: dict):
		"""Initialize the service.
		Args:
			name (str): The name of the service.
			config (dict): The system configuration.
			services (dict, optional): The services to use. Defaults to {}.
		"""
		# Check if the service exists
		if name in services:
			raise ValueError(f'The service {name} already exists.')

		# Set a default configuration and services
		if config is None:
			config = {}
		if services is None:
			services = {}

		# Service information
		self._name = name
		self._global_config = config
		self._config = config[name] if name in config else {}
		self._services = services
		self._manager = None
		self._loop_delay = self.LOOP_DELAY

		# Service status
		self.__initialized = False
		self.__init_try = 0
		self.__errors = 0

		# Service thread
		self.__thread = None
		self.__thread_stop_event = threading.Event()

		# Set the service
		services[name] = self

	@abc.abstractmethod
	def init(self):
		"""Initialize the service."""
		raise NotImplementedError('The init method must be implemented.')

	@abc.abstractmethod
	def destroy(self):
		"""Destroy the service."""
		raise NotImplementedError('The destroy method must be implemented.')

	def start(self):
		"""Start the service."""

		# Check if the service has been initialized
		if self.__initialized:
			return

		# Check if all dependencies are initialized
		for dependency in self.DEPENDENCIES:
			if dependency not in self._services:
				raise ValueError(f'The service {self._name} requires the service {dependency}.')
			if not self._services[dependency].is_initialized():
				log(f'The {self._name} service is waiting for the {dependency} service to start...')
				self.__start_other_service()
				return

		# Initialize the service
		log(f'Starting the {self._name} service...')
		self.init()

		# Call the thread method
		if hasattr(self, 'before') or hasattr(self, 'loop'):
			self.__thread_stop_event.clear()
			self.__thread = threading.Thread(target=self.__thread_execution)
			self.__thread.start()

		# Set the service as initialized
		self.__initialized = True
		log(f'The {self._name} service has started successfully.')

		# Start other services
		services = {k: v for k, v in self._services.items() if not v.is_initialized()}
		if services:
			self.__start_other_service()

	def stop(self):
		"""Stop the service."""
		log(f'Stopping the {self._name} service...')
		if self.__thread:
			self.__thread_stop_event.set()
			if threading.current_thread() != self.__thread:
				self.__thread.join()
		self.destroy()
		self.__initialized = False
		self.__init_try = 0
		log(f'The {self._name} service has stopped successfully.')

	def is_initialized(self) -> bool:
		"""Check if the service has been initialized.
		Returns:
			bool: True if the service has been initialized, False otherwise.
		"""
		return self.__initialized

	def __reboot(self):
		"""Reboot the robot for Linux systems, stop all services and start the service again for other systems."""
		log('Stopping all services...')
		for service in self._services.values():
			service.stop()
		log('All services have been stopped.')
		self.start()

	def __start_other_service(self):
		"""Start another service by try priority."""
		services = {k: v for k, v in self._services.items() if not v.is_initialized()}
		services = dict(sorted(services.items(), key=lambda item: item[1].__init_try))
		service = list(services.values())[0]
		self.__init_try += 1
		service.start()

	def __thread_execution(self):
		"""Service thread execution."""
		# Call the before method
		if hasattr(self, 'before'):
			self.before()

		# Call the loop method
		if hasattr(self, 'loop'):
			if self._global_config.get('system', {}).get('debug', True):
				while not self.__thread_stop_event.is_set():
					self.loop()
					if self.__thread_stop_event.wait(self._loop_delay):
						break
			else:
				while True:
					try:
						self.loop()
						if self.__thread_stop_event.wait(self._loop_delay):
							break
					except KeyboardInterrupt:
						break
					except Exception as e:
						self.__errors += 1
						log(f'An error occurred in the {self._name} service thread: {e}', logging.ERROR)
						if self.MAX_ERRORS_REBOOT != -1 and self.__errors >= self.MAX_ERRORS_REBOOT:
							log(f'The {self._name} service has reached the maximum number of errors. Rebooting the robot...', logging.ERROR)
							self.__reboot()
							break
						if self.MAX_ERRORS != -1 and self.__errors >= self.MAX_ERRORS:
							log(f'The {self._name} service has reached the maximum number of errors. Stopping the service...', logging.ERROR)
							break
						time.sleep(self.ERROR_INTERVAL)



from setuptools import setup, find_packages
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "LICENSE"), "r", encoding="utf-8") as fh:
    license_text = fh.read()

setup(
    name="eymos",
    version="1.1.0",
    author="EymoLabs",
    author_email="info@dzin.es",
    description="EymOS - Lightweight Middleware for Robotics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EymoLabs/eymos",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "opencv-python>=4.5.3.56",
        "Pillow>=8.0.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    license="Custom License",
    keywords="robotics middleware eymos lightweight",
    project_urls={
        "Bug Reports": "https://github.com/EymoLabs/eymos/issues",
        "Source": "https://github.com/EymoLabs/eymos"
    },
)
